import logging

from brownie import chain, ZERO_ADDRESS
from cachetools.func import ttl_cache
from ypricemagic.exceptions import UnsupportedNetwork
from ypricemagic.networks import Network
from ypricemagic.utils.contracts import Contract, Singleton
from ypricemagic.price_modules.chainlink.feeds import FEEDS
from ypricemagic.utils.events import decode_logs, get_logs_asap

logger = logging.getLogger(__name__)

DENOMINATIONS = {
    'ETH': '0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE',
    'BTC': '0xbBbBBBBbbBBBbbbBbbBbbbbBBbBbbbbBbBbbBBbB',
    'USD': '0x0000000000000000000000000000000000000348',
}

registries = {
    # https://docs.chain.link/docs/feed-registry/#contract-addresses
    Network.Mainnet: '0x47Fb2585D2C56Fe188D0E6ec628a38b74fCeeeDf',
}


class Chainlink(metaclass=Singleton):
    def __init__(self):
        if chain.id not in registries and len(FEEDS) == 0:
            raise UnsupportedNetwork('chainlink is not supported on this network')

        if chain.id in registries: self.registry = Contract(registries[chain.id])
        
        self.load_feeds()

    def load_feeds(self):
        if chain.id in registries:
            logs = decode_logs(
                get_logs_asap(str(self.registry), [self.registry.topics['FeedConfirmed']])
            )
            self.feeds = {
                log['asset']: log['latestAggregator']
                for log in logs
                if log['denomination'] == DENOMINATIONS['USD']
            }
        else: self.feeds = {}
        # for mainnet, we have some extra feeds to pull in
        # for non-mainnet, we have no registry so must get feeds manually
        self.feeds.update(FEEDS)
        logger.info(f'loaded {len(self.feeds)} feeds')

    def get_feed(self, asset):
        return Contract(self.feeds[asset])

    def __contains__(self, asset):
        return asset in self.feeds

    @ttl_cache(maxsize=None, ttl=600)
    def get_price(self, asset, block=None):
        if asset == ZERO_ADDRESS:
            return None
        try:
            price = self.get_feed(asset).latestAnswer(block_identifier=block) / 1e8
            logger.debug("chainlink -> %s", price)
            return price
        except ValueError:
            return None


chainlink = None
try:
    chainlink = Chainlink()
except UnsupportedNetwork:
    pass



from typing import Any, Callable, List, Tuple, Union

import brownie
from eth_typing.evm import Address, BlockNumber
from ypricemagic.utils.multicall import multicall_decimals, multicall_totalSupply
from ypricemagic.utils.raw_calls import _decimals, _totalSupply

import y

SUPPORTED_INPUT_TYPES = str, Address, brownie.Contract, y.Contract


def decimals(
    contract_address_or_addresses: Union[str, Address, brownie.Contract, y.Contract, List[Union[str, Address, brownie.Contract, y.Contract]], Tuple[Union[str, Address, brownie.Contract, y.Contract]]],
    block: Union[BlockNumber, int, None] = None, 
    return_None_on_failure: bool = False
    ): 

    func = _choose_appropriate_fn(contract_address_or_addresses, _decimals, multicall_decimals)
    return func(contract_address_or_addresses, block=block, return_None_on_failure=return_None_on_failure)


def totalSupply(
    contract_address_or_addresses: Union[str, Address, brownie.Contract, y.Contract, List[Union[str, Address, brownie.Contract, y.Contract]], Tuple[Union[str, Address, brownie.Contract, y.Contract]]],
    block: Union[BlockNumber, int, None] = None, 
    return_None_on_failure: bool = False
    ):

    func = _choose_appropriate_fn(contract_address_or_addresses, _totalSupply, multicall_totalSupply)
    return func(contract_address_or_addresses, block=block, return_None_on_failure=return_None_on_failure)


def totalSupplyReadable(
    contract_address_or_addresses: Union[str, Address, brownie.Contract, y.Contract, List[Union[str, Address, brownie.Contract, y.Contract]], Tuple[Union[str, Address, brownie.Contract, y.Contract]]],
    block: Union[BlockNumber, int, None] = None, 
    return_None_on_failure: bool = False
    ):

    token_supplys = totalSupply(contract_address_or_addresses, block=block, return_None_on_failure=return_None_on_failure)
    token_decimals = decimals(contract_address_or_addresses, block=block, return_None_on_failure=return_None_on_failure)
    return [supply / 10 ** decimal for supply, decimal in zip(token_supplys, token_decimals)]


def _choose_appropriate_fn(
    input: Any,
    singlecall_fn: Callable,
    multicall_fn: Callable
    ):

    input_type = _input_type(input)
    if input_type == 'single': return singlecall_fn
    elif input_type == 'multi': return multicall_fn


def _input_type(
    input: Any
    ) -> str:

    if type(input) in [list, Tuple]: 
        for value in input: _check_if_supported(value)
        return 'multi'
    
    else:
        _check_if_supported(input)
        return 'single'


def _check_if_supported(
    input: Any
    ) -> bool:
    
    if type(input) not in SUPPORTED_INPUT_TYPES:
        raise TypeError(f'Unsupported input type: {type(input)}')




from brownie import chain
from ypricemagic.networks import Network
from ypricemagic.utils.raw_calls import raw_call


POOLS = {
    Network.BinanceSmartChain: {
        "0x86aFa7ff694Ab8C985b79733745662760e454169": "0xF16D312d119c13dD27fD0dC814b0bCdcaAa62dfD", # Belt.fi bDAI/bUSDC/bUSDT/bBUSD
        "0x9cb73F20164e399958261c289Eb5F9846f4D1404": "0xAEA4f7dcd172997947809CE6F12018a6D5c1E8b6", # 4Belt
    },
}.get(chain.id, {})


def is_belt_lp(token_address):
    return token_address in POOLS


def get_price(token_address, block=None):
    pool = POOLS[token_address]
    return raw_call(pool, 'get_virtual_price()', output='int', block=block) / 1e18




from brownie import chain
from ypricemagic.networks import Network
from ypricemagic.utils.contracts import Contract
from ypricemagic.utils.raw_calls import raw_call, _decimals
from ypricemagic.utils.multicall import fetch_multicall


class NotUniswapPoolV2(Exception):
    pass


class UniswapPoolV2:
    def __init__(self, pool_address: str) -> None:
        self.address = pool_address
        self.decimals = _decimals(self.address)
        self.scale = 10 ** self.decimals
        try: self.factory = raw_call(self.address, 'factory()', output='address')
        except ValueError as e:
            okay_errors = ['is not a valid ETH address','invalid opcode']
            if 'execution reverted' in str(e): raise NotUniswapPoolV2
            elif any([msg in str(e) for msg in okay_errors]):
                try: self.factory = Contract(self.address).factory()
                except AttributeError: raise NotUniswapPoolV2
            else: raise

    def get_pool_details(self, block=None):
        pair = Contract(self.address)
        if chain.id in [Network.Mainnet, Network.BinanceSmartChain, Network.Polygon, Network.Fantom]: 
            token0, token1, supply, reserves = fetch_multicall(
                [pair, "token0"],
                [pair, "token1"],
                [pair, "totalSupply"],
                [pair, "getReserves"],
                block=block
            )
        else: # if your chain does not support multicall
            token0 = pair.token0(block_identifier = block)
            token1 = pair.token1(block_identifier = block)
            supply = pair.totalSupply(block_identifier = block)
            reserves = pair.getReserves(block_identifier = block)

        return token0, token1, supply, reserves


    
    



import logging

from brownie import chain
from ypricemagic.constants import weth, wbtc
from ypricemagic.networks import Network
from ypricemagic.price_modules import yearn
from ypricemagic.utils.raw_calls import _symbol, raw_call

logger = logging.getLogger(__name__)

# This module is far from perfect, but provides an acceptable way to validate some of the prices returned by `get_price`

ACCEPTABLE_HIGH_PRICES = {
    Network.Mainnet: [
        weth.address,
        wbtc.address,
    ],
    Network.BinanceSmartChain: [
        weth.address,
        wbtc.address,
    ],
    Network.Polygon: [
        weth.address,
        wbtc.address,
    ],
    Network.Fantom: [
        weth.address,
        wbtc.address,
    ],
}.get(chain.id, [])

def _sense_check(
    token_address: str, 
    price: float
    ):

    if yearn.is_yearn_vault(token_address):
        token_address = raw_call(token_address, 'token()', output='address')

    if price > 1000 and token_address not in ACCEPTABLE_HIGH_PRICES:
        price_readable = round(price, 4)
        symbol = _symbol(token_address)
        network = Network.name(chain.id)
        logger.warn(f'unusually high price (${price_readable}) returned for {symbol} on {network}. This does not necessarily mean that the price is wrong, but you may want to validate the price for yourself before proceeding.')
